"""
Custom loss function for ECGs.

MB Vessies and RR van de Leur
"""

import torch
import torch.nn.functional as F
import pytorch_lightning as pl
from typing import Union, List
import operator
from operator import itemgetter
import numpy as np
from ecgxai.utils.exceptions import SumMeanReductionShapeMismatchException


class BaseLoss(pl.LightningModule):
    """
    Base class, used for inheritance later on.

    Attributes:
        reduction: String defining how to reduced the computed loss per sample 
        on the batch level. Options are: 'mean', 'sum' and 'sum_mean'. This last 
        option only works if the input is three dimension, and uses a sum for the 
        second and third dimension, while the a mean over the batch dimension is taken.
    """

    def __init__(self, reduction : str = 'mean'):
        super().__init__()
        assert reduction in ['sum', 'mean', 'sum_mean'], "Please provide one of the following for the reduction function: 'mean', 'sum' of 'sum_mean'."

        self.reduction = reduction
        self.only_log_num = True

    @property
    def input_args(self):
        raise NotImplementedError("Make sure to define the input_args property in each loss class")

    @property
    def batch_args(self):
        return []

    def apply_reduction(self, loss, skip_if_none : bool = False):
        """
        Reduces the computed loss per sample on the batch level.

        Args:
            loss: The computed loss per sample.
            skip_if_none: TODO
        """

        if self.reduction == 'sum':
            return loss.sum()
        elif self.reduction == 'mean':
            return loss.mean()
        elif self.reduction == 'sum_mean':
            if len(loss.shape) != 3:
                raise SumMeanReductionShapeMismatchException(dim=len(loss.shape))
            return loss.sum(dim=(1, 2)).mean()
        # TODO, must be a better way to handle this
        # skip_if_none is generally controlled by the only_log_num attribute to make sure no N-dimensional tensors are logged
        return loss if not skip_if_none else None


class TW(BaseLoss):
    """ 
    This wrapper can convert any standard Pytorch loss into a class
    that is usable in the ECGx.AI package.

    Attributes:
        loss: A PyTorch loss object that needs to be converted.
        input_args: A list of strings containing the input arguments
            that are needed to compute the loss. This differs per 
            loss function, and could be `x` and `y` for the 
            CrossEntropyLoss or `x` and `reconstruction` for the MSE in an auto-encoder.
        batch_args: A list of strings containing the arguments that are
            needed from the the batch dictionary, for example P on- and offsets
            to calculate locations specific losses.
    """
    def __init__(
        self,
        loss : object,
        input_args : List[str],
        batch_args : List[str] = [],
        long_args : List[str] = [],
        arg_max_args : List[str] = [],
        **kwargs
    ):
        super().__init__(**kwargs)

        self.inp_args = input_args
        self.b_args = batch_args
        self.long_args = long_args
        self.arg_max_args = arg_max_args
        self.loss = loss

    @property
    def input_args(self):
        return self.inp_args

    @property
    def batch_args(self):
        return self.b_args

    def __call__(self, args):
        l_args = {k: v for k, v in args.items() if k in self.b_args}
        l_args.update({k: v for k, v in args.items() if k in self.inp_args})

        # Take argmax of last dim of args if specified
        for k in self.arg_max_args:
            l_args[k] = l_args[k].argmax(dim=-1)

        # Cast args to int if specified
        for k in self.long_args:
            l_args[k] = l_args[k].long()

        loss = self.loss(*l_args.values())
        return self.apply_reduction(loss), {}


class CombinedLoss(BaseLoss):
    """ 
    Combines mulitple losses into a single loss class.
    
    Attributes:
        losses: List of loss functions (defined as objects) to combine
        operations: List of operations, or single operation to used to combine the loss functions
            defined as strings. Currently supported are "+", "-", "*" and "/". 
            If a list is supplied it should be the same lenght as losses.
        combine_weigths: Weights to multiply each loss by before applying opperations (optional).
    """
    def __init__(
        self,
        losses : List[object],
        operations : Union[List[str], str] = '+',
        combine_weigths : List[float] = None,
        **kwargs
    ):
        """
        Initializes the combined loss.
        """
        super().__init__(**kwargs)

        self.losses = torch.nn.Sequential(*losses)

        # Dict so that we can change the strings to callable functions
        opp_dict = {
            "+": operator.add,
            "-": operator.sub,
            "*": operator.mul,
            "/": operator.truediv
        }
                
        if isinstance(operations, list):
            assert len(operations) == len(self.losses) - 1, "Operations should be a string or a list of strings of length len(losses) - 1"
        else:
            # If only one operator is given, apply same operation to each loss
            operations = [operations for _ in range(len(self.losses) - 1)]

        # convert operation strings to callable functions
        try:
            self.operations = [opp_dict[opp] for opp in operations] 
        except KeyError:
            KeyError('Undfined operation, supported operations are are "+", "-", "*" and "/".')

        # Add none as first operation, as the zip function later will not work without one
        self.operations.insert(0, None)

        # check length of the given weigths, otherwise just set all the weights to one
        if combine_weigths:
            assert len(combine_weigths) == len(self.losses), "Losses and weights should be of equal length"
            self.combine_weigths = combine_weigths
        else:
            self.combine_weigths = torch.ones(len(self.losses))

        # Allow sub-losses to return 'none' reduced logs (loss and log reduction are controlled by the same parameter), logs are then reduced by combined loss class
        for loss in self.losses:
            loss.only_log_num = False

    @property
    def batch_args(self):
        return {arg for loss in self.losses for arg in loss.batch_args}

    @property
    def input_args(self):
        return {arg for loss in self.losses for arg in loss.input_args}

    def __call__(self, args : dict):
        """
        Internal function to compute the combined loss.

        Args:
            args: Dictionary with possible loss arguments indexed by loss parameter name
        """
        logs = {}

        for idx, (loss_obj, operation, weight) in enumerate(zip(self.losses, self.operations, self.combine_weigths)):
            # Call the loss function with the corresponding arguments and multiply it by correspoinding weight
            try:
                c_loss, c_logs = loss_obj(args)
                weighted_c_loss = weight * c_loss

                logs[f'{type(loss_obj).__name__}'] = self.apply_reduction(c_loss, skip_if_none=self.only_log_num)
                logs[f'{type(loss_obj).__name__}_weighted'] = self.apply_reduction(weighted_c_loss, skip_if_none=self.only_log_num)

                logs.update({f'{type(loss_obj).__name__}_{k}': self.apply_reduction(v, skip_if_none=self.only_log_num) for k, v in c_logs.items()})

            except KeyError as e:
                raise KeyError(f"Failed to fetch loss arguments {e.args} for {type(loss_obj).__name__}, available arguments are {args.keys()}.")

            # Combine previous loss and current loss using set operation if applicable (which is not true for first loss)
            loss = weighted_c_loss if idx == 0 else operation(loss, weighted_c_loss)

        return self.apply_reduction(loss), logs


class KLDivergence(BaseLoss):
    """
    Computes the Kuback-Leibner divergence for training of variational auto-encoders.
    
    Attributes:
        std_is_log: If the standard deviation is provided to the loss function as ln(SD).
        mu_min_clip: Minimum value to clip the mean on.
        mu_max_clip: Maximum value to clip the mean on.
        std_min_clip: Minimum value to clip the standard deviation on.
        std_max_clip: Maximum value to clip the standard deviation on.
    """
    def __init__(
        self,
        std_is_log : bool = True,
        mu_min_clip : float = -100,
        mu_max_clip : float = 100,
        std_min_clip : float = 1e-6,
        std_max_clip : float = 100,
        **kwargs
    ):
        super().__init__(**kwargs)
        self.std_is_log = std_is_log
        self.mu_min_clip = mu_min_clip
        self.mu_max_clip = mu_max_clip
        self.std_min_clip = std_min_clip
        self.std_max_clip = std_max_clip

    @property
    def input_args(self):
        return ['mu', 'std']

    def __call__(self, args):
        mu, std = args['mu'], args['std']
        logs = {}
        if self.std_is_log:
            std = torch.exp(std)

        # clip std and mu
        mu = mu.clip(min=self.mu_min_clip, max=self.mu_max_clip)
        std = std.clip(min=self.std_min_clip , max=self.std_max_clip)

        loss = -0.5 * torch.sum(1 + std.pow(2).log() - mu.pow(2) - std.pow(2), dim=1)

        return self.apply_reduction(loss), logs


class GaussianVAEReconLoss(BaseLoss):
    """
    Computes the Gaussian Reconstruction Loss as proposed by Van de Leur & Bos et al (2022).

    For this loss to work, your model should return both a mean and standard deviation for
    every timepoint in the ECG.

    Attributes:
        recon_nan_clip_val: Value to replace NaN values in the reconstruction mean or SD with.
    """

    def __init__(
        self,
        recon_nan_clip_val : float = 50,
        **kwargs
    ):
        super().__init__(**kwargs)
        self.recon_nan_clip_val = recon_nan_clip_val

    @property
    def input_args(self):
        return ['x', 'reconstruction_mean', 'reconstruction_std']

    def __call__(self, args):
        x, reconstruction_mean, reconstruction_std = itemgetter(*self.input_args)(args)
        logs = {}

        # set nan to self.recon_nan_clip_val
        # TODO check why we need this, add it to other VAE system?
        reconstruction_mean[reconstruction_mean != reconstruction_mean] = self.recon_nan_clip_val
        reconstruction_std[reconstruction_std != reconstruction_std] = self.recon_nan_clip_val

        x = x.flatten(start_dim=1)
        reconstruction_mean = reconstruction_mean.flatten(start_dim=1)
        reconstruction_std = reconstruction_std.flatten(start_dim=1)

        loss = -torch.sum(
            -(0.5 * np.log(2 * np.pi) + 0.5 * reconstruction_std.pow(2).log())
            - 0.5 * ((x - reconstruction_mean).pow(2) / reconstruction_std.pow(2)), dim=1)

        return self.apply_reduction(loss), logs


class MSELoss(BaseLoss):
    """
    Computes the mean squared error (MSE) loss between the original ECG and its reconstruction.
    """
    def __init__(
        self,
        **kwargs
    ):
        super().__init__(**kwargs)

    @property
    def input_args(self):
        return ['x', 'reconstruction']
    
    def __call__(self, args):
        logs = {}
        x, reconstruction = args['x'], args['reconstruction']
        loss = (x - reconstruction)**2
        return self.apply_reduction(loss), logs


class BinaryFocalLoss(BaseLoss):
    def __init__(self, gamma=2, alpha=1, pos_weight=None):
        """Initializes the binary version of the focal loss.

        Args:
            gamma (int): The gamma value for the focal loss.
            alpha (int): The alpha value for the focal loss.
            pos_weight (tensor): The weight given to the positive class
                to deal with the class imbalance even more.
        """
        super(BinaryFocalLoss, self).__init__()
        self.gamma = gamma
        self.alpha = alpha
        self.pos_weight = pos_weight if pos_weight is not None else pos_weight

    @property
    def input_args(self):
        return ['label', 'y_hat']
        
    def forward(self, args):
        """Calculate the loss."""
        y, y_hat = args['label'], args['y_hat'].squeeze()
        logs = {}
        
        pos_weight = self.pos_weight
        BCE_loss = F.binary_cross_entropy_with_logits(
            y_hat, y, reduction='none', pos_weight=pos_weight)
        # prevents nans when probability 0
        pt = torch.exp(-BCE_loss)
        focal_loss = self.alpha * (1 - pt)**self.gamma * BCE_loss

        return self.apply_reduction(focal_loss), logs



"""
Autoencoder systems.

DB van de Pavert, RR van de Leur and MB Vessies
"""

import torch.nn as nn

from ecgxai.utils.exceptions import ModelOutputShapeMismatchException
from ecgxai.systems.base_system import BaseSystem


class AE(BaseSystem):
    def __init__(
        self,
        encoder_class: nn.Module,
        decoder_class: nn.Module,
        **kwargs
    ):
        super().__init__(**kwargs)

        self.save_hyperparameters()

        self.encoder = encoder_class
        self.decoder = decoder_class

    def forward(self, batch):
        x = batch['waveform']
        z = self.encoder(x)
        reconstruction = self.decoder(z)

        if x.shape != reconstruction.shape:
            raise ModelOutputShapeMismatchException(reconstruction.shape, x.shape)

        return reconstruction, z

    @property
    def step_args(self):
        return ['x', 'reconstruction', 'z']

    def _step(self, batch):
        x = batch['waveform']
        reconstruction, z = self.forward(batch)

        if x.shape != reconstruction.shape:
            raise ModelOutputShapeMismatchException(reconstruction.shape, x.shape)

        return {
            'x': x,
            'reconstruction': reconstruction,
            'z': z
        }




"""
Exceptions.

MB Vessies
"""

class ModelOutputShapeMismatchException(Exception):
    def __init__(self, shape : tuple = None, correct_shape : tuple = None, message : str = None):
        if not message:
            message = f'Model output shape {shape} does not match expected output shape {correct_shape}.'
        super().__init__(message)


class SumMeanReductionShapeMismatchException(Exception):
    def __init__(self, dim : int = None, message : str = None):
        if not message:
            message = f'Loss reduction "sum_mean" requires a 3 dimenstional output, instead an output of shape {dim} was given.'
        super().__init__(message)


class IncorrectNumberOfTMWLabelsException(Exception):
    def __init__(self, metric : object, expected : int, recieved : int, message : str = None):
        if not message:
            message = f'Wrapped metric of type {type(metric).__name__} has an output of size {expected} but {recieved} labels were given.'
        super().__init__(message)



"""
Base system

MB Vessies
"""

import pytorch_lightning as pl
import torch.optim as optim
from typing import List
from torchmetrics import MetricCollection
import re
import torch


class BaseSystem(pl.LightningModule):
    def __init__(
        self, loss : object,
        lr: float,
        train_metrics : MetricCollection = None,
        val_metrics : MetricCollection = None,
        test_metrics : MetricCollection = None,
        log_filter : List[str] = None,
        log_filter_whitelist : bool = False
    ):
        super().__init__()

        self.__check_loss(loss)

        self.train_metrics = train_metrics.clone(prefix='train_') if train_metrics else None
        self.valid_metrics = val_metrics.clone(prefix='val_') if val_metrics else None
        self.test_metrics = test_metrics.clone(prefix='test_') if test_metrics else None

        self.__check_metrics_input(train_metrics, prefix="(train)")
        self.__check_metrics_input(val_metrics, prefix="(val)")
        self.__check_metrics_input(test_metrics, prefix="(test)")

        self.loss = loss
        self.lr = lr
        self.log_filter = log_filter
        self.log_filter_whitelist = log_filter_whitelist

    def __check_loss(self, loss):
        print(loss.input_args)
        for loss_arg in loss.input_args:
            if loss_arg not in self.step_args:
                raise KeyError(f"Loss of type '{type(loss).__name__}' requires input_arg {loss_arg} but it is not defined as a return argument of the current system")

    def __check_metrics_input(self, metrics, prefix=""):
        k_map = {
            '(train)': self.train_metrics,
            '(val)': self.valid_metrics,
            '(test)': self.test_metrics
        }

        if not k_map[prefix]:
            return

        for metric_name, metric in metrics.items():
            for metric_arg in metric.input_args:
                if metric_arg not in self.step_args:
                    raise KeyError(f"{prefix} Metric '{metric_name}' of type {type(metric).__name__} requires input_arg {metric_arg} but it is not defined as a returned argument of the current system")

    def __retrieve_loss_batch_args(self, batch, step_dict):
        for loss_arg in self.loss.batch_args:
            if loss_arg not in batch.keys():
                raise KeyError(f"Loss '{type(self.loss).__name__}' requires batch_arg {loss_arg} but it is not found in the given batch")
            step_dict[loss_arg] = batch[loss_arg]
        return step_dict

    def __retrieve_metrics_batch_args(self, metrics, batch, step_dict, prefix=""):
        if not metrics:
            return step_dict

        for metric_name, metric in metrics.items():
            for metric_arg in metric.batch_args:
                if metric_arg not in batch.keys():
                    raise KeyError(f"{prefix} Metric '{metric_name}' of type {type(metric).__name__} requires batch_arg {metric_arg} but it is not found in the given batch")
                step_dict[metric_arg] = batch[metric_arg]
        return step_dict

    def __apply_log_filter(self, log_dict):
        if self.log_filter:
            res = {}
            for filt in self.log_filter:
                for k, v in log_dict.items():
                    does_match = bool(re.search(filt, k))

                    if not (does_match ^ self.log_filter_whitelist):
                        res[k] = v
            return res
        else:
            return log_dict

    def configure_optimizers(self):
        """ TODO make configureable """

        self.optimizer = optim.Adam(self.parameters(), lr=self.lr)
        return {"optimizer": self.optimizer}

    @property
    def step_args(self):
        raise NotImplementedError("Make sure to define the step_args property in each system")

    def _step(self, batch):
        raise NotImplementedError

    def forward(self, batch):
        raise NotImplementedError

    def training_step(self, batch, batch_idx):
        step_dict = self._step(batch)
        step_dict = self.__retrieve_loss_batch_args(batch, step_dict)

        loss, logs = self.loss(step_dict)
        logs = self.__apply_log_filter(logs)
        self.log_dict({f"train_{k}": v for k, v in logs.items()})
        if self.train_metrics:
            step_dict = self.__retrieve_metrics_batch_args(self.train_metrics, batch, step_dict, prefix="(train)")
            metric_results = self.train_metrics(step_dict)
            self.log_dict(metric_results, on_step=True, on_epoch=True)

        self.log("train_loss", loss)
        return loss

    def validation_step(self, batch, batch_idx):
        step_dict = self._step(batch)
        step_dict = self.__retrieve_loss_batch_args(batch, step_dict)

        loss, logs = self.loss(step_dict)
        logs = self.__apply_log_filter(logs)
        self.log_dict({f"val_{k}": v for k, v in logs.items()})

        if self.valid_metrics:
            step_dict = self.__retrieve_metrics_batch_args(self.valid_metrics, batch, step_dict, prefix="(val)")
            metric_results = self.valid_metrics(step_dict)
            self.log_dict(metric_results, on_step=False, on_epoch=True)

        self.log("val_loss", loss)

    def test_step(self, batch, batch_idx):
        step_dict = self._step(batch)
        step_dict = self.__retrieve_loss_batch_args(batch, step_dict)
        
        loss, logs = self.loss(step_dict)
        logs = self.__apply_log_filter(logs)
        self.log_dict({f"test_{k}": v for k, v in logs.items()})

        if self.test_metrics:
            step_dict = self.__retrieve_metrics_batch_args(self.test_metrics, batch, step_dict, prefix="(test)")
            metric_results = self.test_metrics(step_dict)
            step_dict.update(metric_results)
            self.log_dict(metric_results, on_step=False, on_epoch=True)

        self.log("test_loss", loss)

        return step_dict

    def predict_step(self, batch, batch_idx):
        return self._step(batch)



import torch
import torch.nn as nn
from .modules import CausalCNN, Softplus

class CausalCNNVDecoder(torch.nn.Module):
    """
    Variational decoder.
    """
    def __init__(self, k, width, in_channels, channels, depth, out_channels,
                 kernel_size, gaussian_out, softplus_eps, dropout):
        super(CausalCNNVDecoder, self).__init__()
        self.in_channels = in_channels
        self.width = width
        self.gaussian_out = gaussian_out
        self.linear1 = torch.nn.Linear(k, in_channels)
        self.linear2 = torch.nn.Linear(in_channels, in_channels * width)
        self.causal_cnn = CausalCNN(
            in_channels, channels, depth, out_channels, kernel_size,
            forward=False,
        )
        if self.gaussian_out:
            self.linear_mean = nn.Linear(out_channels * width, 
                                         out_channels * width)
            self.linear_sd = torch.nn.Sequential(
                torch.nn.Linear(out_channels * width, out_channels * width),
                Softplus(softplus_eps),
            )
        
    def forward(self, x):
        """
        Returns a reconstruction of the original 8x600 ECG, by decoding
        the given compression.
        """
        B, _ = x.shape
        # from (BxK) to (BxC)
        out = self.linear1(x)
        # from (BxC) to (Bx(C*600))
        out = self.linear2(out)
        # from (Bx(C*600)) to (BxCx600)
        out = out.view(B, self.in_channels, self.width)
        # deconvolve through the causal CNN
        out = self.causal_cnn(out)
        if self.gaussian_out:
            nflat_shape = out.shape
            # flatten the output to shape (Bx(8*600))
            out = torch.flatten(out, start_dim=1)
            return self.linear_mean(out).reshape(nflat_shape), self.linear_sd(out).reshape(nflat_shape)
        return out


"""
Variational autoencoder system

MB Vessies and RR van de Leur
"""

import torch
from torch import nn

from ecgxai.utils.exceptions import ModelOutputShapeMismatchException
from ecgxai.systems.base_system import BaseSystem


class VAE(BaseSystem):
    """ Basic VAE system """

    def __init__(
        self,
        encoder_class: nn.Module,
        decoder_class: nn.Module,
        std_is_log : bool = True,
        **kwargs
    ):
        super(VAE, self).__init__(**kwargs)
        
        self.save_hyperparameters()

        self.std_is_log = std_is_log

        self.encoder = encoder_class
        self.decoder = decoder_class

    def reparameterize(self, mu : torch.Tensor, std : torch.Tensor):
        """
            The 'reparameterization trick', allowing for differentiable sampling
            Arguments:
                mu : torch.Tensor       -- Mean of the normal distribution to sample from
                std : torch.Tensor      -- Standard deviation of the distribution to sample from, if input is log_std, this should
                    be specified through the 'std_is_log' parameter of the system
            returns:
                z : torch.Tensor        -- Sample from the distribution
        """
        if self.std_is_log:
            std = std.exp()
        eps = torch.normal(torch.zeros_like(mu), torch.ones_like(std))
        z = mu + (eps * std)
        return z

    def forward(self, x : torch.Tensor, deterministic : bool = False):
        """
            Takes a batch dictionary of ECGs as input and reconstucts it through the VAE model

            Arguments:
                x : dict                -- (batch of) ECG(s) used as input for the model
                deterministic : bool    -- Wether or not apply the reperameterization trick and use a sampling operation in the latent space, 
                    might be preferable to disable this during inference
            Returns:
                reconstruction : torch.Tensor -- Reconstruction of the orginal input
        """
        if len(x.shape) == 2:
            x = torch.unsqueeze(x, dim=0)
            
        mu, std = self.encoder(x)
        z = mu if deterministic else self.reparameterize(mu, std)

        reconstruction = self.decoder(z)

        if x.shape != reconstruction.shape:
            raise ModelOutputShapeMismatchException(reconstruction.shape, x.shape)

        return {
            'x': x,
            'reconstruction': reconstruction,
            'z': z,
            'mu': mu,
            'std': std,
        }

    @property
    def step_args(self):
        return ['x', 'reconstruction', 'mu', 'std']
    
    def _step(self, batch : dict):
        """
            Takes an ECG as input and reconstucts it through the VAE model, then calculates the loss for training

            Arguments:
                batch : dict     -- (batch dictionary of) ECG(s) used as input for the model, ECG waveform is expected under key ['waveform'] in the dict
            Returns:
                loss : float     --  Loss
                step_dict : dict --  Dictionary of intermediate calculation results as well as final model outputs
        """
        x = batch['waveform']

        return self.forward(x)


class GaussianVAE(VAE):
    """ VAE system with mean and standard deviation as final output """

    def __init__(
        self,
        **kwargs
    ):
        super(GaussianVAE, self).__init__(**kwargs)

    def forward(self, x : torch.Tensor, deterministic : bool = False):
        """
            Takes a batch of ECGs as input and reconstucts it through the VAE model

            Arguments:
                x : tensor              -- (batch of) ECG(s) used as input for the model
                deterministic : bool    -- Wether or not apply the reperameterization trick and use a sampling operation in the latent space,
                    might be preferable to disable this during inference
            Returns:
                results : dict          -- Dictionary containing all outputs of the VAE
        """
        if len(x.shape) == 2:
            x = torch.unsqueeze(x, dim=0)

        mu, std = self.encoder(x)
        z = mu if deterministic else self.reparameterize(mu, std)

        reconstruction_mean, reconstruction_std = self.decoder(z)

        if x.shape != reconstruction_mean.shape:
            raise ModelOutputShapeMismatchException(message=f'Model output (mean) shape {x.shape} does not match expected output shape {reconstruction_mean.shape}.')

        if x.shape != reconstruction_std.shape:
            raise ModelOutputShapeMismatchException(message=f'Model output (std) shape {x.shape} does not match expected output shape {reconstruction_std.shape}.')

        return {
            'x': x,
            'reconstruction': reconstruction_mean,
            'reconstruction_mean': reconstruction_mean,
            'reconstruction_std': reconstruction_std,
            'z': z,
            'mu': mu,
            'std': std,
        }

    @property
    def step_args(self):
        return ['x', 'reconstruction', 'reconstruction_mean', 'reconstruction_std', 'mu', 'std']  

    def _step(self, batch : dict):
        """
            Takes an ECG as input and reconstucts it through the VAE model, then calculates the loss for training

            Arguments:
                batch : dict     -- (batch dictionary of) ECG(s) used as input for the model, ECG waveform is expected under key ['waveform'] in the dict
            Returns:
                step_dict : dict --  Dictionary of intermediate calculation results as well as final model outputs
        """
        x = batch['waveform']

        return self.forward(x)



import torch
import torch.nn as nn
from .encoder import CausalCNNVEncoder
from .decoder import CausalCNNVDecoder


class VAE(nn.Module):
    """
    The VAE method is able to encode a given input into
    mean and log. variance parameters for Gaussian
    distributions that make up a compressed space that represents
    the input. Then, a sample from that space can be decoded into
    an attempted, probably less detailed, reconstruction of the
    original input.
    """
    
    def __init__(self, encoder_params, decoder_params):
        super().__init__()
        self.encoder_params = encoder_params
        self.decoder_params = decoder_params
        self.encoder = CausalCNNVEncoder(**encoder_params)
        self.decoder = CausalCNNVDecoder(**decoder_params)
    
    def reparameterize(self, mu, sd):
        """Sample from a Gaussian distribution with given mean and s.d."""
        eps = torch.randn_like(sd)
        return mu + eps * sd
    
    def forward(self, x):
        """Return reconstructed input after compressing it"""
        enc_mu, enc_sd = self.encoder(x)
        z = self.reparameterize(enc_mu, enc_sd)
        if self.decoder_params['gaussian_out']:
            dec_mu, dec_sd = self.decoder(z)
            # reconstruction is equal to the mean value for the gauss. distr. of each point
            # reshape the mean vector to be of same size as input (Bx8x600 for median ECGs)
            recon_x = dec_mu.view(x.shape)
            return recon_x, z, [(enc_mu, enc_sd), (dec_mu, dec_sd)]
        recon_x = self.decoder(z)
        return recon_x, z, [(enc_mu, enc_sd)]



"""
Various transformations used for data preprocessing

RR van de Leur, MN Bos and MB Vessies
"""

import torch
import numpy as np
from scipy import interpolate


class ToTensor(object):
    """
    Converts the ECG waveform and label numpy arrays to FloatTensors.
    """
    def __call__(self, sample):
        waveform = sample['waveform']
        secondary_waveform = None
        if 'secondary_waveform' in sample:
            secondary_waveform = sample['secondary_waveform']    

        sample['waveform'] = torch.from_numpy(waveform).type(torch.float)
        if secondary_waveform is not None:
            sample['secondary_waveform'] = torch.from_numpy(secondary_waveform).type(torch.float)

        if 'label' in sample:
            sample['label'] = torch.from_numpy(
                np.array(sample['label'])).type(torch.float)

        return sample


class ApplyGain(object):
    """
    Applies the gain to the ECG signal to convert it into millivolts. The millivolt
    values of the ECG are usually in a good range for neural networks, so no
    other normalization is needed. For the physionet dataset, a correction for the 
    baseline is also performed.
    """
    def __call__(self, sample):
        sample['waveform'] = sample['waveform'] * sample['gain']
        baseline_corrections = [(int(k.split('_')[1]), v) for k, v in sample.items() if 'TrueBaseline_' in k]
        for lead_idx, correction in baseline_corrections:
            sample['waveform'][lead_idx] -= correction

        return sample


class To12Lead(object):
    """
    Convert 8 lead ECGs (with I, II, V1-V6 in this order) to their 12 lead equivalent.
    """
    def _to12lead(self, waveform):
        out = torch.zeros((12, waveform.shape[1]))
        # I and II
        out[0:2, :] = waveform[0:2, :] 
        # III = II - I
        out[2, :] = waveform[1, :] - waveform[0, :]
        # aVR = -(I + II)/2
        out[3, :] = -(waveform[0, :] + waveform[1, :]) / 2
        # aVL = I - II/2
        out[4, :] = waveform[0, :] - (waveform[1, :] / 2)
        # aVF = II - I/2
        out[5, :] = waveform[1, :] - (waveform[0, :] / 2)
        # V1 to V6
        out[6:12, :] = waveform[2:8, :]

        return out

    def __call__(self, sample):
        waveform = sample['waveform']
        
        assert waveform.shape[0] == 8, "The To12Lead transform only works with 8 channel input ECGs, please check."

        sample['waveform'] = self._to12lead(waveform)
        
        secondary_waveform = None
        if 'secondary_waveform' in sample:
            secondary_waveform = sample['secondary_waveform']
            sample['secondary_waveform'] = self._to12lead(secondary_waveform)

        return sample


class Resample(object):
    """Resamples the ECG to the specified sampling frequency.

    Attributes:
        sample_freq: The required sampling frequency to resample to.
    """
    def __init__(self, sample_freq):
        """Initializes the resample transformation.
        """
        self.sample_freq = int(sample_freq)

    def _resample(self, waveform, samplebase):
        length = int(waveform.shape[1])
        out_length = int((length / samplebase) * self.sample_freq)

        if type(waveform) == np.ndarray:
            x = np.linspace(0, length / samplebase, num=length)
            f = interpolate.interp1d(x, waveform, axis=1)
            xnew = np.linspace(0, length / samplebase, num=out_length)
            return f(xnew)
        elif type(waveform) == torch.Tensor:
            return torch.nn.functional.interpolate(waveform.unsqueeze(0), size=out_length, mode='linear', align_corners=False).squeeze(0)
                  
    def __call__(self, sample):
        samplebase = int(sample['samplebase'])
        waveform = sample['waveform']
        secondary_waveform = None
        if 'secondary_waveform' in sample:
            secondary_waveform = sample['secondary_waveform']

        if samplebase != self.sample_freq:
            sample['waveform'] = self._resample(waveform, samplebase)

            # Do the same for secondary waveform
            if secondary_waveform is not None:
                sample['secondary_waveform'] = self._resample(secondary_waveform, samplebase)

        return sample


class PolyFilter(object):
    """Filter ECG by subtracting a polynomial.

    Attributes:
        order (int): The order of the filter, usually 2, 3 or 4.
    """
    def __init__(self, order):
        """Initializes the resample transformation.
        """
        self.order = int(order)

    def __call__(self, sample):
        waveform = sample['waveform']
        ecg_out = np.zeros_like(waveform)

        for lead in range(waveform.shape[0]):
            z = np.polyfit(np.linspace(0, 10, waveform.shape[1]), waveform[lead, :], self.order)
            yp = np.poly1d(z)
            ecg_out[lead, :] = waveform[lead, :] - yp(np.linspace(0, 10, waveform.shape[1]))

        sample['waveform'] = ecg_out

        return sample




import torch
import torch.nn as nn
import pytorch_lightning as pl


class CNNDoubleResidual(pl.LightningModule):
    """
        Residual 'double' CNN as described by  Hannun et al. converting a multi channel input to some flat output dimension
    """
    def __init__(
            self,
            num_layers: int = 15,
            in_sample_dim: int = 600,
            in_channels: int = 32,
            kernel_size: int = 16,
            dropout_rate: float = 0.2,
            sub_sample_every: int = 4,
            double_channel_every: int = 4,
            act_func : nn.Module = nn.ReLU(),
            batchnorm : bool = True
    ):
        super(CNNDoubleResidual, self).__init__()
        
        self.num_layers = num_layers
        self.in_sample_dim = in_sample_dim
        self.in_channels = in_channels
        self.sub_sample_every = sub_sample_every
        self.double_channel_every = double_channel_every

        self.layers = []

        layer_sample_dim = in_sample_dim
        for layer_idx in range(num_layers):
            # Double output channels every 4th layer
            output_channels = self.in_channels * (2**((layer_idx + 1) // self.double_channel_every))
            # Subsample input every few layers
            sub_sample_rate = 2 if layer_idx > 0 and layer_idx % sub_sample_every == 0 else 1

            if int(layer_sample_dim + 1) % 2 == 0 and bool(kernel_size % 2):
                sub_sample_extra_pad = -1
            elif int(layer_sample_dim + 1) % 2 == 0 or bool(kernel_size % 2):
                sub_sample_extra_pad = 0
            else:
                sub_sample_extra_pad = 1

            self.layers.append(
                ResidualMaxPoolDoubleConvBlockForward(
                    in_channels,
                    output_channels,
                    sub_sample_rate=sub_sample_rate,
                    kernel_size=kernel_size,
                    dropout_rate=dropout_rate,
                    batchnorm=batchnorm,
                    act_func=act_func,
                    sub_sample_extra_pad=sub_sample_extra_pad
                )
            )
            layer_sample_dim /= sub_sample_rate
            in_channels = output_channels

        # print(self.calculate_output_dim())

        self.layers.append(nn.Flatten())
        self.layers = nn.Sequential(*self.layers)

    def calculate_output_dim(self):
        cnn_output_channels = self.in_channels * (2**((self.num_layers) // self.double_channel_every))
        cnn_output_samples = self.in_sample_dim // (2**((self.num_layers - 1) // self.sub_sample_every))
        return int(cnn_output_channels * cnn_output_samples), cnn_output_channels, cnn_output_samples

    def forward(self, x):
        return self.layers(x)


class ResidualMaxPoolDoubleConvBlockForward(nn.Module):
    """
        Residual convblock adding a max-pool over the input to the out, performs 2 convolutions 
        with (optionally) a batchnorm an activation function and dropout.
    """

    def __init__(
        self,
        in_dim,
        out_dim,
        sub_sample_rate: int = 2,
        kernel_size: int = 16,
        dropout_rate : float = 0.2,
        act_func : nn.Module = nn.ReLU(),
        batchnorm : bool = True,
        sub_sample_extra_pad : int = 1
    ):
        super(ResidualMaxPoolDoubleConvBlockForward, self).__init__()

        self.sub_sample_rate = sub_sample_rate
        self.channel_padding = out_dim - in_dim
        
        self.pool = nn.MaxPool1d(self.sub_sample_rate)

        # assuming keras style 'same' padding -- padding is applied through nn.ConstantPad1d
        P = int((kernel_size - 1) // 2)
        padding = (P, P + 1 if kernel_size % 2 == 0 else P)

        pre_modules = []
        if batchnorm:
            pre_modules.append(nn.BatchNorm1d(in_dim))

        if act_func:
            pre_modules.append(act_func)

        self.f = nn.Sequential(
            *pre_modules,
            nn.ConstantPad1d(padding, 0),
            nn.Conv1d(in_dim, out_dim, kernel_size, stride=self.sub_sample_rate),
            nn.BatchNorm1d(out_dim),
            nn.ReLU(),
            nn.Dropout(p=dropout_rate),
            nn.ConstantPad1d((P, P + sub_sample_extra_pad) if self.sub_sample_rate % 2 == 0 else padding, 0),
            nn.Conv1d(out_dim, out_dim, kernel_size, stride=1),
            nn.BatchNorm1d(out_dim)
        )

    def forward(self, x):
        return self.pool(torch.functional.F.pad(x, (0, 0, 0, self.channel_padding))) + self.f.forward(x)


class Reshape(nn.Module): 
    """ Reshapes an input tensor into the given shape """
    def __init__(self, *args):
        super(Reshape, self).__init__()
        self.shape = args

    def forward(self, x):
        return x.view(self.shape)



"""
Binary classificatios sytem

DB van de Pavert and MB Vessies
"""

import torch
import torch.nn as nn

from ecgxai.systems.base_system import BaseSystem


class ClassificationSystem(BaseSystem):
    def __init__(
        self,
        model: nn.Module,
        mode: str,
        **kwargs
    ):
        super().__init__(**kwargs)

        assert mode in ['binary', 'multi_class', 'multi_label'], 'mode should be either "binary", "multi_class" or "multi_label"'  

        self.prob_func = nn.Softmax(dim=-1) if mode == 'multi_class' else torch.sigmoid
        
        self.save_hyperparameters()

        self.model = model

    def forward(self, x : torch.Tensor):
        y = self.model(x)
        return y

    @property
    def step_args(self):
        return ['y_hat', 'y_prob', 'label']

    def _step(self, batch):
        x = batch['waveform']
        y_hat = self.forward(x)
        return {
            'y_hat': y_hat,
            'y_prob': self.prob_func(y_hat),
            'label': batch['label']
        }


"""
ECG datasets to use with ECGx.AI package.

MB Vessies and RR van de Leur
"""

import os
import torch
from torch.utils.data import Dataset
from scipy import io
from sklearn.model_selection import train_test_split
import numpy as np
from typing import List, Union, Callable
import pandas as pd
from torchvision.transforms import Compose
from torchvision.datasets.utils import download_url
from tqdm import tqdm
import subprocess


class DatasetBase(Dataset):
    """
        Only used to inherit some functions later on.
    """
    def get_dataframe(self):
        """
        Returns the dataframe that is being used in this dataset.
        """
        return self.dataset

    def query_dataset(
        self,
        query : List[tuple],
        max_num_samples : int = None,
        return_indices : bool = False
    ):
        """
        Query dataset for specific class (combinations), returns samples (dict with wavefrom etc) or indices in dataset.

        Args:
            query: Defines what classes returned samples should (not) have e.g. [('rhythm_st', True), ('conduction_lbtb': False)]
            max_num_samples: Maximum number of samples to return from query (default: {None})
            return_indices: Wether to return samples (dict with waveform etc) or indices in dataset (default: {False})

        Returns:
            List of samples or list of indices that meet the query.
        """
        query = query if type(query) == list else [query]

        qs = " and ".join([f'{param} == {val}' for (param, val) in query])
        qres = self.dataset.query(qs)
        if max_num_samples and max_num_samples < qres.shape[0]:
            qres = qres[:max_num_samples]
        
        if return_indices:
            return [self.dataset.index.get_loc(qidx) for qidx in qres.index]

        return [self.__getitem__(self.dataset.index.get_loc(qidx)) for qidx in qres.index]

    def print_stats(self):
        """ Prints statistics of the dataset. """
        stats, full_size = self.get_stats()
        print('-- Dataset distribution -- ')
        print(f'Full size: {full_size}')
        for s in stats:
            print(f'["{s["class"]}"] - Num entries: {s["size"]} ({s["fraction"]:.3}%)')

    def __len__(self):
        """ Returns the length of the dataset. """
        return len(self.dataset)


class UniversalECGDataset(DatasetBase):
    """
        Universal ECG dataset in PyTorch Dataset format.

        This class defines a dataset that can be used to train PyTorch models
        using a wide range of different ECG formats. On initialization, this
        dataset format can be defined, or a custom function to load the data
        can be provided. During training, a dictionary is returned on every
        iteration with the following contents:
        
        {
            'waveform': numpy array with the ECG waveform
            'samplebase': int of the sampling frequency,
            'gain': float of the gain of the ECG, to convert it into millivolts,
            'id': the id of the sample,
        }

        Optionally, some other variables are also provided, when they exist
        in the dataframe. These can be used for transformations or loss functions
        later on.

        Attributes:
            dataset_function: Selects the function to use to load dataset samples. Predefined 
                options are 'umcu', 'universal', 'physionet' and 'physionet_numpy'. In the 
                UMCU datasets, ECGs are saved in the numpy format using the following folder structure:
                `waveform_dir/ps/eu/doid/testid.npy`. In the universal datasets, files are also saved using
                the numpy format, but then all in one folder: `waveform_dir/filename.npy`. For the physionet
                datasets, files are saved either using the matlab format of the numpy format. It is also
                possible to provide your own function to load custom datasets. Inside this function you 
                will have access to all columns in the provided dataset to correctly load your
                custom ECG file. Make sure you return the ECG with the channels or leads as the first dimension
                and the samples as the second dimension.
            waveform_dir: Path of the folder with the raw waveform files are stored.
            dataset: Pandas DataFrame with the dataset your are using. For all datasets, the columns
                SampleBase (the sampling frequency) and the Gain (to convert the signal to mV) are required.
                The UMCU dataset requires the PseudoID and TestID columns, while all other datasets
                only require a Filename column.
            transform: List of PyTorch transformations.
            labels: Name(s) of the y variable in the dataset, 
                if a list of names is supplied a vector is returned.
            secondary_waveform_dir: Directory of a possible secondary waveform, can be used if multiple waveforms
                from different locations are required for a model to train. E.g. we can use it when we try 
                to convert rhythm ECGs to median beat ECGs. To use the secondary waveform, the dataset should 
                be saved in the same format as the main dataset. 
    """
    def __init__(
        self, 
        dataset_function: Union[str, Callable],
        waveform_dir: str,
        dataset: pd.DataFrame,
        transform: Union[Compose, None] = None,
        labels: Union[str, List[str]] = 'Label',
        secondary_waveform_dir: str = ""
    ):
        """
        Initializes the ECG datasets.
        """
        assert dataset_function in ['umcu', 'universal', 'physionet', 'physionet_numpy'] or isinstance(dataset_function, Callable)

        if dataset_function == "umcu":
            if not ('PseudoID' in dataset and 'TestID' in dataset):
                raise ValueError("Please provide a dataframe with both PseudoID and TestID columns when using the UMCU format.")
        elif dataset_function in ['universal', 'physionet', 'physionet_numpy']:
            if 'Filename' not in dataset:
                raise ValueError("Please provide a dataframe with Filename column when using this format.")

        self.dataset = dataset
        self.waveform_dir = waveform_dir
        self.secondary_waveform_dir = secondary_waveform_dir
        self.transform = transform
        self.labels = labels
        self.label_indices = [self.dataset.columns.get_loc(c) for c in labels] if isinstance(labels, list) else None
        self.dataset_function = dataset_function

    def train_test_split(self, ratio=0.1, shuffle=True):
        """
            Splits the current dataset into a train and testset. 
        """
        trainset_df, testset_df = train_test_split(self.dataset, test_size=ratio, shuffle=shuffle)

        trainset = UniversalECGDataset(
            self.dataset_function,
            self.waveform_dir,
            trainset_df,
            self.transform,
            self.labels,
            self.secondary_waveform_dir,
        )

        testset = UniversalECGDataset(
            self.dataset_function,
            self.waveform_dir,
            testset_df,
            self.transform,
            self.labels,
            self.secondary_waveform_dir,
        )

        return trainset, testset

    @staticmethod
    def loadUMCUSample(data_dir: str, pseudo_id: str, test_id: str):
        """
        Loads the raw ECG data stored in the UMCU format into an numpy array.

        ECGs are saved in the numpy format using the following folder structure:
        `waveform_dir/ps/eu/doid/testid.npy`.

        Args:
            data_dir: Directory where the ECGs are stored.
            pseudo_id: The ID denoting the patient or subject.
            test_id: The ID denoting the individual recording.

        Returns:
            A numpy array with the ECG voltage data.
        """
        return np.load(os.path.join(data_dir, pseudo_id[0:2], pseudo_id[2:4], pseudo_id[4:], f'{test_id}.npy',)), test_id

    @staticmethod
    def loadUniversalSample(data_dir: str, filename: str):
        """
        Loads the raw ECG data stored in the universal format into an numpy array. 
        
        ECGs are saved in the numpy format using the following folder structure:
        `waveform_dir/filename.npy`.

        Args:
            data_dir: Directory where the ECGs are stored.
            filename: The name of the ECG file.

        Returns:
            A numpy array with the ECG voltage data.
        """
        return np.load(os.path.join(data_dir, f'{filename}.npy',)), filename

    @staticmethod
    def loadPhysionetSample(data_dir: str, filename: str):
        """
        Loads the raw ECG data stored in the physionet matlab format into an numpy array. 
        
        ECGs are saved in the numpy format using the following folder structure:
        `waveform_dir/filename.mat`. The raw ECG data is saved in the 'val' variable in Matlab.

        Args:
            data_dir: Directory where the ECGs are stored.
            filename: The name of the ECG file.

        Returns:
            A numpy array with the ECG voltage data.
        """
        path = PhysioNetDataset.split_to_path(filename, steps=3)
        return io.loadmat(os.path.join(data_dir, path, f'{filename}.mat',))['val'], filename

    @staticmethod
    def loadPhysionetNumpySample(data_dir: str, filename: str):
        """
        Loads the raw ECG data stored in the physionet numpy format into an numpy array. 
        
        ECGs are saved in the numpy format using the following folder structure:
        `waveform_dir/filename.npy`.

        Args:
            data_dir: Directory where the ECGs are stored.
            filename: The name of the ECG file.

        Returns:
            A numpy array with the ECG voltage data.
        """
        path = PhysioNetDataset.split_to_path(filename, steps=3)
        return np.load(os.path.join(data_dir, path, f'{filename}.npy')), filename

    def get_stats(self):
        """
        Calculates size and fraction (percentage) of each label in dataset 

        Returns:
            Tuple with a list of dictonaries with 
            {'class': [classname], 'size': [num samples in class], 'fraction': [fraction of total dataset]},
            and total dataset size.
        """
        full_size = len(self.dataset.index)
        stats = []
        labels = self.labels if isinstance(self.labels, list) else [self.labels]
        for label in labels:
            class_size = len(self.dataset[self.dataset[label] == 1].index)
            stats.append({'class': label, 'size': class_size, 'fraction': (class_size / full_size) * 100})

        return stats, full_size

    def _load_waveform(self, idx):
        """
        Internal function to load the ECG waveform from a file in the prespecified format.
        """
        secondary_waveform = None
        if isinstance(self.dataset_function, Callable):
            # List will be passed through *, meaning the only argument will be the df row
            args = [self.dataset.iloc[idx]]
            wf_load_function = self.dataset_function
        else:
            if self.dataset_function == 'umcu':
                args = [self.dataset['PseudoID'].iloc[idx], self.dataset['TestID'].iloc[idx]]
                wf_load_function = self.loadUMCUSample

            elif self.dataset_function == 'universal':
                args = [self.dataset['Filename'].iloc[idx]]
                wf_load_function = self.loadUniversalSample

            elif self.dataset_function == 'physionet':
                args = [self.dataset['Filename'].iloc[idx]]
                wf_load_function = self.loadPhysionetSample

            elif self.dataset_function == 'physionet_numpy':
                args = [self.dataset['Filename'].iloc[idx]]
                wf_load_function = self.loadPhysionetNumpySample

        waveform, sample_id = wf_load_function(self.waveform_dir, *args)
        if self.secondary_waveform_dir:
            secondary_waveform, _ = wf_load_function(self.secondary_waveform_dir, *args)

        return sample_id, waveform, secondary_waveform

    def __getitem__(self, idx):
        "Internal function to return an ECG sample."
        sample_id, waveform, secondary_waveform = self._load_waveform(idx)

        # Add waveform, original sample base, gain and ID to sample
        sample = {
            'waveform': waveform,
            'samplebase': int(self.dataset['SampleBase'].iloc[idx]),
            'gain': float(self.dataset['Gain'].iloc[idx]),
            'id': sample_id,
        }

        if secondary_waveform is not None:
            sample['secondary_waveform'] = secondary_waveform

        # Sometimes additional information is needed (e.g. for a median cutoff)
        possible_cols = ['AcqDate', 'POnset', 'TOffset', 'VentricularRate',
                         'QOnset', 'POffset', 'QOffset', 'start_idx',
                         'end_idx'] + [f'TrueBaseline_{i}' for i in range(12)]

        for col in possible_cols:
            if col in self.dataset:
                sample[col.lower()] = self.dataset[col].iloc[idx]

        if self.labels:
            if isinstance(self.labels, list):
                labels = self.dataset.iloc[idx, self.label_indices].astype('int64')
                sample['label'] = torch.from_numpy(labels.values)
            elif self.labels in self.dataset.columns.values:
                label = self.dataset[self.labels].iloc[idx]
                sample['label'] = label

        try:
            if self.transform:
                # for now always applies the same transforms to secondary sample
                sample = self.transform(sample)
        except Exception as e:
            print(e)
            print("Above error was caught in dataloader, returning neighbouring sample to continue training")
            idx = idx - 1 if idx > 0 else idx + 1
            return self.__getitem__(idx)

        return sample


class PhysioNetDataset(UniversalECGDataset):
    @staticmethod
    def split_to_path(filename, steps=3):
        """
         Generates path a path where every 2 characters from the input filename form a directory
        """
        return os.path.join(*[filename[i:i + 2] for i in range(steps)])

    # TODO make property or constant somewhere else
    @ staticmethod
    def SNOMEDCT_To_Abbreviation_map():
        return {'164889003': 'AF', '164890007': 'AFL', '6374002': 'BBB', '426627000': 'Brady', '733534002': 'CLBBB', '713427006': 'CRBBB', '270492004': 'IAVB', '713426002': 'IRBBB', '39732003': 'LAD', '445118002': 'LAnFB', '164909002': 'LBBB', '251146004': 'LQRSV', '698252002': 'NSIVCB', '426783006': 'NSR', '284470004': 'PAC', '10370003': 'PR', '365413008': 'PRWP', '427172004': 'PVC', '164947007': 'LPR', '111975006': 'LQT', '164917005': 'QAb', '47665007': 'RAD', '59118001': 'RBBB', '427393009': 'SA', '426177001': 'SB', '427084000': 'STach', '63593006': 'SVPB', '164934002': 'TAb', '59931005': 'TInv', '17338001': 'VPB'}

    @ staticmethod
    def extract_header_info(file_content, file_id):
        _, n_leads, sample_base, num_samples, date, time = file_content[0].split()

        info = {
            'Filename': file_id,
            'n_leads': int(n_leads),
            'SampleBase': int(sample_base),
            'NumSamples': int(num_samples),
            'Date': date,
            'Time': time,
            'AF': 0,
            'AFL': 0,
            'BBB': 0,
            'Brady': 0,
            'CLBBB': 0,
            'CRBBB': 0,
            'IAVB': 0,
            'IRBBB': 0,
            'LAD': 0,
            'LAnFB': 0,
            'LBBB': 0,
            'LQRSV': 0,
            'NSIVCB': 0,
            'NSR': 0,
            'PAC': 0,
            'PR': 0,
            'PRWP': 0,
            'PVC': 0,
            'LPR': 0,
            'LQT': 0,
            'QAb': 0,
            'RAD': 0,
            'RBBB': 0,
            'SA': 0,
            'SB': 0,
            'STach': 0,
            'SVPB': 0,
            'TAb': 0,
            'TInv': 0,
            'VPB': 0
        }

        if n_leads != '12':
            raise RuntimeError(f"Found non 12-lead entry, {file_id}")

        sr = set()
        baseline_correction = {}
        extra_fields = ['#Rx:', '#Hx:', '#Sx:']
        try:
            for idx, l in enumerate(file_content[1:]):
                l = l.split()
                if l[0][-4:] == '.mat':
                    baseline_correction[f'TrueBaseline_{idx}'] = int(l[4])
                    sr.add(l[2])
                elif l[0] == '#Age:':
                    info['Age'] = float(l[-1])
                elif l[0] == '#Sex:':
                    info['Sex'] = PhysioNetDataset.convert_sex(l[-1])
                elif l[0] == '#Dx:':
                    for diag in l[-1].split(','):
                        if diag in PhysioNetDataset.SNOMEDCT_To_Abbreviation_map().keys():
                            diag = PhysioNetDataset.SNOMEDCT_To_Abbreviation_map()[diag]
                            info[diag] = 1
                elif l[0] in extra_fields:
                    info[l[0][1:]] = l[-1]
        
            if len(sr) > 1:
                raise RuntimeError(f"Found inconsistend gain per lead in {file_id}")

            info.update(baseline_correction)
            info['Gain'] = float(sr.pop().split("/")[0]) * 1e-6

        except Exception as e:
            print(file_content)
            raise e

        return info

    @staticmethod
    def convert_sex(sex):
        if sex in ('Female', 'female', 'F', 'f'):
            return 0
        elif sex in ('Male', 'male', 'M', 'm'):
            return 1
        return float('nan')

    def convert_to_numpy(self, df, path):
        for filename in tqdm(df['Filename']):
            file_path = PhysioNetDataset.split_to_path(filename, steps=3)
            if not os.path.exists(os.path.join(path, file_path, f'{filename}.npy')):
                data = io.loadmat(os.path.join(path, file_path, f'{filename}.mat',))['val']
                np.save(os.path.join(path, file_path, f'{filename}.npy'), data)

    def extract_archive(
        self,
        remove_finished: bool = False,
    ):
        if self.to_path is None:
            self.to_path = os.path.dirname(self.from_path)

        print("Now extracting data from tar archive, this may take a while...")

        process = subprocess.Popen(f'tar zxf {self.from_path} -C {self.to_path}', shell=True)
        process.wait()

        print(f"Done extracting - exit code = {process.returncode}")

        print("Now extracting header info and moving files")
        header_info = []

        for file in os.listdir(f'{self.to_path}/{self.archive_root}'):
            full_file_path = f'{self.to_path}/{self.archive_root}/{file}'
            file_id, ftype = file.split('.')

            # Only read header files, extract raw data in header file is found
            if ftype == 'hea':
                with open(full_file_path, 'r') as extracted:
                    # Get header info from header file
                    header_info.append(self.extract_header_info(extracted.readlines(), file_id))

                mat_file_name = f'{self.to_path}/{self.archive_root}/{file_id}.mat'

                # Create folder structure for raw data
                file_path = PhysioNetDataset.split_to_path(file_id, steps=3)
                f_to_path = os.path.join(self.to_path, file_path)

                # Move files
                os.renames(mat_file_name, os.path.join(f_to_path, f'{file_id}.mat'))
                os.renames(full_file_path, os.path.join(f_to_path, f'{file_id}.hea'))

        header_df = pd.DataFrame(header_info)
        header_df.to_csv(f"{self.to_path}/header_info.csv")

        print("All done")

        if remove_finished:
            os.remove(self.from_path)


class PTBXLDataset(PhysioNetDataset):
    def __init__(self, path : str = "./PTB_XL", download : bool = True, use_numpy: bool = True, *args, **kwargs):
        self.download = download
        self.use_numpy = use_numpy

        if download:
            download_url(
                'https://storage.googleapis.com/physionetchallenge2021-public-datasets/WFDB_PTBXL.tar.gz', 
                path, 
                filename='PTB_XL.tar.gz',
                md5='55e8a5c25eadfeff4fcd38f5bbf3cb13'
            )

        self.archive_root = 'WFDB_PTBXL'
        self.from_path = f'{path}/PTB_XL.tar.gz'
        self.to_path = None

        # Assume extraction was done correctly if header_info csv exists
        # TODO, is this enough? 
        if not os.path.exists(f"{path}/header_info.csv"):
            self.extract_archive()

        df = pd.read_csv(f"{path}/header_info.csv")

        if use_numpy:
            self.convert_to_numpy(df, path)

        super(PTBXLDataset, self).__init__(
            dataset_function='physionet_numpy' if use_numpy else 'physionet',
            waveform_dir=path,
            dataset=df,
            *args,
            **kwargs
        )


class CPSC2018Dataset(PhysioNetDataset):
    def __init__(self, path : str = "./CPSC_2018", download : bool = True, use_numpy: bool = True, *args, **kwargs):
        self.download = download
        self.use_numpy = use_numpy

        if download:
            download_url(
                'https://storage.googleapis.com/physionetchallenge2021-public-datasets/WFDB_CPSC2018.tar.gz', 
                path, 
                filename='CPSC_2018.tar.gz',
                md5='5d4b520e3b6558a33dc9dbe49d08f8f1'
            )

        self.archive_root = 'WFDB_CPSC2018'
        self.from_path = f'{path}/CPSC_2018.tar.gz'
        self.to_path = None

        # Assume extraction was done correctly if header_info csv exists
        # TODO, is this enough? 
        if not os.path.exists(f"{path}/header_info.csv"):
            self.extract_archive()

        df = pd.read_csv(f"{path}/header_info.csv")

        if use_numpy:
            self.convert_to_numpy(df, path)

        super(CPSC2018Dataset, self).__init__(
            dataset_function='physionet_numpy' if use_numpy else 'physionet',
            waveform_dir=path,
            dataset=df,
            *args,
            **kwargs
        )


class China12LeadDataset(PhysioNetDataset):
    def __init__(self, path : str = "./China_12_Lead", download : bool = True, use_numpy: bool = True, *args, **kwargs):
        self.download = download
        self.use_numpy = use_numpy

        if download:
            download_url(
                'https://storage.googleapis.com/physionetchallenge2021-public-datasets/WFDB_CPSC2018_2.tar.gz', 
                path, 
                filename='China_12_Lead.tar.gz',
                md5='5b1498abacaa1b5a762691c006e737ad'
            )

        self.archive_root = 'WFDB_CPSC2018_2'
        self.from_path = f'{path}/China_12_Lead.tar.gz'
        self.to_path = None

        # Assume extraction was done correctly if header_info csv exists
        # TODO, is this enough? 
        if not os.path.exists(f"{path}/header_info.csv"):
            self.extract_archive()

        df = pd.read_csv(f"{path}/header_info.csv")

        if use_numpy:
            self.convert_to_numpy(df, path)

        super(China12LeadDataset, self).__init__(
            dataset_function='physionet_numpy' if use_numpy else 'physionet',
            waveform_dir=path,
            dataset=df,
            *args,
            **kwargs
        )


class INCARTDataset(PhysioNetDataset):
    def __init__(self, path : str = "./INCART", download : bool = True, use_numpy: bool = True, *args, **kwargs):
        self.download = download
        self.use_numpy = use_numpy

        if download:
            download_url(
                'https://storage.googleapis.com/physionetchallenge2021-public-datasets/WFDB_StPetersburg.tar.gz', 
                path, 
                filename='INCART.tar.gz',
                md5='525dde6bd26bff0dcb35189e78ae7d6d'
            )

        self.archive_root = 'WFDB_StPetersburg'
        self.from_path = f'{path}/INCART.tar.gz'
        self.to_path = None

        # Assume extraction was done correctly if header_info csv exists
        # TODO, is this enough? 
        if not os.path.exists(f"{path}/header_info.csv"):
            self.extract_archive()

        df = pd.read_csv(f"{path}/header_info.csv")

        if use_numpy:
            self.convert_to_numpy(df, path)

        super(INCARTDataset, self).__init__(
            dataset_function='physionet_numpy' if use_numpy else 'physionet',
            waveform_dir=path,
            dataset=df,
            *args,
            **kwargs
        )


class PTBDataset(PhysioNetDataset):
    def __init__(self, path : str = "./PTB", download : bool = True, use_numpy: bool = True, *args, **kwargs):
        self.download = download
        self.use_numpy = use_numpy

        if download:
            download_url(
                'https://storage.googleapis.com/physionetchallenge2021-public-datasets/WFDB_PTB.tar.gz', 
                path, 
                filename='PTB.tar.gz',
                md5='3df4662a8a9189a6a5924424b0fcde0e'
            )

        self.archive_root = 'WFDB_PTB'
        self.from_path = f'{path}/PTB.tar.gz'
        self.to_path = None

        # Assume extraction was done correctly if header_info csv exists
        # TODO, is this enough? 
        if not os.path.exists(f"{path}/header_info.csv"):
            self.extract_archive()

        df = pd.read_csv(f"{path}/header_info.csv")

        if use_numpy:
            self.convert_to_numpy(df, path)

        super(PTBDataset, self).__init__(
            dataset_function='physionet_numpy' if use_numpy else 'physionet',
            waveform_dir=path,
            dataset=df,
            *args,
            **kwargs
        )


class Georgia12LeadDataset(PhysioNetDataset):
    def __init__(self, path : str = "./Georgia_12_Lead", download : bool = True, use_numpy: bool = True, *args, **kwargs):
        self.download = download
        self.use_numpy = use_numpy

        if download:
            download_url(
                'https://storage.googleapis.com/physionetchallenge2021-public-datasets/WFDB_Ga.tar.gz', 
                path, 
                filename='Georgia_12_Lead.tar.gz',
                md5='47085dd62baca5ace4041025d6910b13'
            )

        self.archive_root = 'WFDB_Ga'
        self.from_path = f'{path}/Georgia_12_Lead.tar.gz'
        self.to_path = None

        # Assume extraction was done correctly if header_info csv exists
        # TODO, is this enough? 
        if not os.path.exists(f"{path}/header_info.csv"):
            self.extract_archive()

        df = pd.read_csv(f"{path}/header_info.csv")

        if use_numpy:
            self.convert_to_numpy(df, path)

        super(Georgia12LeadDataset, self).__init__(
            dataset_function='physionet_numpy' if use_numpy else 'physionet',
            waveform_dir=path,
            dataset=df,
            *args,
            **kwargs
        )


class ChapmanUniversityDataset(PhysioNetDataset):
    def __init__(self, path : str = "./Chapman_University", download : bool = True, use_numpy: bool = True, *args, **kwargs):
        self.download = download
        self.use_numpy = use_numpy

        if download:
            download_url(
                'https://storage.googleapis.com/physionetchallenge2021-public-datasets/WFDB_ChapmanShaoxing.tar.gz', 
                path, 
                filename='Chapman_University.tar.gz',
                md5='a3e10171eba1e7520a38919594d834e5'
            )

        self.archive_root = 'WFDB_ChapmanShaoxing'
        self.from_path = f'{path}/Chapman_University.tar.gz'
        self.to_path = None

        # Assume extraction was done correctly if header_info csv exists
        # TODO, is this enough? 
        if not os.path.exists(f"{path}/header_info.csv"):
            self.extract_archive()

        df = pd.read_csv(f"{path}/header_info.csv")

        if use_numpy:
            self.convert_to_numpy(df, path)

        super(ChapmanUniversityDataset, self).__init__(
            dataset_function='physionet_numpy' if use_numpy else 'physionet',
            waveform_dir=path,
            dataset=df,
            *args,
            **kwargs
        )


class NingboDataset(PhysioNetDataset):
    def __init__(self, path : str = "./Ningbo", download : bool = True, use_numpy: bool = True, *args, **kwargs):
        self.download = download
        self.use_numpy = use_numpy

        if download:
            download_url(
                'https://storage.googleapis.com/physionetchallenge2021-public-datasets/WFDB_Ningbo.tar.gz', 
                path, 
                filename='Ningbo.tar.gz',
                md5='84171145922078146875394acb89b765'
            )

        self.archive_root = 'WFDB_Ningbo'
        self.from_path = f'{path}/Ningbo.tar.gz'
        self.to_path = None

        # Assume extraction was done correctly if header_info csv exists
        # TODO, is this enough? 
        if not os.path.exists(f"{path}/header_info.csv"):
            self.extract_archive()

        df = pd.read_csv(f"{path}/header_info.csv")

        if use_numpy:
            self.convert_to_numpy(df, path)

        super(NingboDataset, self).__init__(
            dataset_function='physionet_numpy' if use_numpy else 'physionet',
            waveform_dir=path,
            dataset=df,
            *args,
            **kwargs
        )


import torch
import torch.nn as nn


class Softplus(nn.Module):
    """
    Applies Softplus to the output and adds a small number.

    Attributes:
        eps (int): Small number to add for stability.
    """
    def __init__(self, eps: float):
        super(Softplus, self).__init__()
        self.eps = eps
        self.softplus = nn.Softplus()

    def forward(self, x):
        return self.softplus(x) + self.eps


class Chomp1d(torch.nn.Module):
    """
    Removes the last elements of a time series.

    Takes as input a three-dimensional tensor (`B`, `C`, `L`) where `B` is the
    batch size, `C` is the number of input channels, and `L` is the length of
    the input. Outputs a three-dimensional tensor (`B`, `C`, `L - s`) where `s`
    is the number of elements to remove.

    Attributes:
        chomp_size (int): Number of elements to remove.
    """
    def __init__(self, chomp_size: int):
        super(Chomp1d, self).__init__()
        self.chomp_size = chomp_size

    def forward(self, x):
        return x[:, :, :-self.chomp_size]


class SqueezeChannels(torch.nn.Module):
    """
    Squeezes, in a three-dimensional tensor, the third dimension.
    """
    def __init__(self):
        super(SqueezeChannels, self).__init__()

    def forward(self, x):
        return x.squeeze(2)


class CausalConvolutionBlock(torch.nn.Module):
    """
    Causal convolution block, composed sequentially of two causal convolutions
    (with leaky ReLU activation functions), and a parallel residual connection.

    Takes as input a three-dimensional tensor (`B`, `C`, `L`) where `B` is the
    batch size, `C` is the number of input channels, and `L` is the length of
    the input. Outputs a three-dimensional tensor (`B`, `C`, `L`).

    Attributes:
        in_channels: Number of input channels.
        out_channels: Number of output channels.
        kernel_size: Kernel size of the applied non-residual convolutions.
        padding: Zero-padding applied to the left of the input of the
           non-residual convolutions.
        final (bool) Disables, if True, the last activation function.
    """
    def __init__(self, in_channels: int, out_channels: int, kernel_size: int,
                 dilation: int, final=False, forward=True):
        super(CausalConvolutionBlock, self).__init__()

        Conv1d = torch.nn.Conv1d if forward else torch.nn.ConvTranspose1d
        
        # Computes left padding so that the applied convolutions are causal
        padding = (kernel_size - 1) * dilation

        # First causal convolution
        conv1 = Conv1d(
            in_channels, out_channels, kernel_size,
            padding=padding, dilation=dilation
        )
        # The truncation makes the convolution causal
        chomp1 = Chomp1d(padding)
        relu1 = torch.nn.LeakyReLU()

        # Second causal convolution
        conv2 = Conv1d(
            out_channels, out_channels, kernel_size,
            padding=padding, dilation=dilation
        )
        chomp2 = Chomp1d(padding)
        relu2 = torch.nn.LeakyReLU()

        # Causal network
        self.causal = torch.nn.Sequential(
            conv1, chomp1, relu1, conv2, chomp2, relu2
        )

        # Residual connection
        self.upordownsample = Conv1d(
            in_channels, out_channels, 1
        ) if in_channels != out_channels else None

        # Final activation function
        self.relu = torch.nn.LeakyReLU() if final else None

    def forward(self, x):
        out_causal = self.causal(x)
        res = x if self.upordownsample is None else self.upordownsample(x)
        if self.relu is None:
            return out_causal + res
        else:
            return self.relu(out_causal + res)


class CausalCNN(torch.nn.Module):
    """
    Causal CNN, composed of a sequence of causal convolution blocks.

    Takes as input a three-dimensional tensor (`B`, `C`, `L`) where `B` is the
    batch size, `C` is the number of input channels, and `L` is the length of
    the input. Outputs a three-dimensional tensor (`B`, `C_out`, `L`).

        in_channels (int): Number of input channels.
        channels (int): Number of channels processed in the network and of output
           channels.
        depth (int): Depth of the network.
        out_channels (int): Number of output channels.
        kernel_size (int): Kernel size of the applied non-residual convolutions.
    """
    def __init__(self, in_channels, channels, depth, out_channels,
                 kernel_size, forward=True):
        super(CausalCNN, self).__init__()

        layers = []  # List of causal convolution blocks
        # double the dilation size if forward, if backward
        # we start at the final dilation and work backwards
        dilation_size = 1 if forward else 2**depth
        
        for i in range(depth):
            in_channels_block = in_channels if i == 0 else channels
            layers += [CausalConvolutionBlock(
                in_channels_block, channels, kernel_size, dilation_size,
                forward,
            )]
            # double the dilation at each step if forward, otherwise
            # halve the dilation
            dilation_size = dilation_size * 2 if forward else dilation_size // 2

        # Last layer
        layers += [CausalConvolutionBlock(
            channels, out_channels, kernel_size, dilation_size
        )]

        self.network = torch.nn.Sequential(*layers)

    def forward(self, x):
        return self.network(x)


class Spatial(nn.Module):
    def __init__(self, channels, dropout, forward=True):
        super(Spatial, self).__init__()
        Conv1d = nn.Conv1d if forward else nn.ConvTranspose1d
        self.network = nn.Sequential(
            Conv1d(channels, channels, 1),
            nn.BatchNorm1d(num_features=channels),
            nn.ReLU(),
            nn.Dropout(dropout),
        )

    def forward(self, x):
        return self.network(x)


import torch
import torch.nn as nn
from .modules import CausalCNN, Softplus, SqueezeChannels


class CausalCNNEncoder(torch.nn.Module):
    """
    Encoder of a time series using a causal CNN: the computed representation is
    the output of a fully connected layer applied to the output of an adaptive
    max pooling layer applied on top of the causal CNN, which reduces the
    length of the time series to a fixed size.

    Takes as input a three-dimensional tensor (`B`, `C`, `L`) where `B` is the
    batch size, `C` is the number of input channels, and `L` is the length of
    the input. Outputs a three-dimensional tensor (`B`, `C`).

    Args:
        in_channels (int): Number of input channels.
        channels (int): Number of channels manipulated in the causal CNN.
        depth (int): Depth of the causal CNN.
        reduced_size (int): Fixed length to which the output time series of the
           causal CNN is reduced.
        out_channels (int): Number of output classes.
        kernel_size (int): Kernel size of the applied non-residual convolutions.
        dropout (float): The dropout probability between 0 and 1.
    """
    def __init__(self, in_channels, channels, depth, reduced_size,
                 out_channels, kernel_size, dropout):
        super(CausalCNNEncoder, self).__init__()
        causal_cnn = CausalCNN(
            in_channels, channels, depth, reduced_size, kernel_size
        )
        spatial = Spatial(reduced_size, dropout)
        reduce_size = torch.nn.AdaptiveAvgPool1d(1)
        squeeze = SqueezeChannels()  # Squeezes the third dimension (time)
        linear1 = torch.nn.Linear(reduced_size, 26)
        linear2 = torch.nn.Linear(26, out_channels)
        self.network = torch.nn.Sequential(
            causal_cnn, spatial, reduce_size, squeeze, linear1,
            nn.BatchNorm1d(num_features=26), nn.ReLU(), nn.Dropout(dropout), linear2,
        )

    def forward(self, x):
        return self.network(x)


class CausalCNNVEncoder(torch.nn.Module):
    """
    Variational encoder. Difference is that we need two outputs: mean and
    standard deviation.

    Args:
        in_channels (int): Number of input channels.
        channels (int): Number of channels manipulated in the causal CNN.
        depth (int): Depth of the causal CNN.
        reduced_size (int): Fixed length to which the output time series of the
           causal CNN is reduced.
        out_channels (int): Number of output classes.
        kernel_size (int): Kernel size of the applied non-residual convolutions.
        softplus_eps (float): Small number to add for stability of the Softplus activation.
        dropout (float): The dropout probability between 0 and 1.
        sd_output (bool): Put to true when using this class inside a VAE, as
            an additional output for the SD is added.
    """
    def __init__(self, in_channels: int, channels: int, depth: int, reduced_size: int,
                 out_channels: int, kernel_size: int, softplus_eps: float, dropout: float, 
                 sd_output: bool = True):
        super(CausalCNNVEncoder, self).__init__()
        causal_cnn = CausalCNN(
            in_channels, channels, depth, reduced_size, kernel_size
        )
        reduce_size = torch.nn.AdaptiveMaxPool1d(1)
        squeeze = SqueezeChannels()  # Squeezes the third dimension (time)
        self.network = torch.nn.Sequential(
            causal_cnn, reduce_size, squeeze,
        )
        self.linear_mean = torch.nn.Linear(reduced_size, out_channels)
        self.sd_output = sd_output
        if self.sd_output:
            self.linear_sd = torch.nn.Sequential(
                torch.nn.Linear(reduced_size, out_channels),
                Softplus(softplus_eps),
            )

    def forward(self, x):
        out = self.network(x)
        if self.sd_output:
            return self.linear_mean(out), self.linear_sd(out)
        return self.linear_mean(out).squeeze()



from setuptools import setup, find_packages

setup(name='ecgxai',
      version='0.1.0',
      description='Neatly packaged AI methods for ECG analysis',
      author='Rutger van de Leur',
      author_email='r.r.vandeleur@umcutrecht.nl',
      license='GNU AGPLv3',
      packages=find_packages(include=['ecgxai', 'ecgxai.*']),
      url="https://github.com/rutgervandeleur/ecgxai",
      install_requires=[
          "pytorch_lightning==1.5.10",
          "torchmetrics==0.9.1",
          "torch==1.9.*",
          "numpy",
          "scipy",
          "pandas",
          "scikit-learn",
          "tqdm"
        ])


import torch
from torchmetrics import Metric
from typing import Union, List
import ecgxai.utils.exceptions as exceptions


class BaseMetric(Metric):
    @property
    def batch_args(self):
        return []

    @property
    def input_args(self):
        raise NotImplementedError()


class TMW(BaseMetric):
    """ Torch Metric wrapper, wrapper class for default TorchMetrics"""
    def __init__(
        self,
        metric : object,
        input_args : List[str],
        batch_args : List[str] = [],
        int_args : List[str] = [],
        out_labels : List[str] = None,
        **kwargs
    ):
        super().__init__(**kwargs)

        self.inp_args = input_args
        self.b_args = batch_args
        self.int_args = int_args
        self.metric = metric
        self.out_labels = out_labels

    @property
    def input_args(self):
        return self.inp_args

    @property
    def batch_args(self):
        return self.b_args

    def update(self, args):
        m_args = {k: v for k, v in args.items() if k in self.b_args}
        m_args.update({k: v for k, v in args.items() if k in self.inp_args})

        # Cast args to int if specified
        for k in self.int_args:
            m_args[k] = m_args[k].int()

        self.metric(*m_args.values())

    def compute(self):
        out = self.metric.compute()

        if not self.out_labels:
            return out

        if len(self.out_labels) != len(out):
            raise exceptions.IncorrectNumberOfTMWLabelsException(self.metric, len(out.values()[0]), len(self.out_labels))

        res = {}
        for label, value in zip(self.out_labels, out):
            res[label] = value.detach().cpu()

        return res

